// Copyright 2019 Diffuse Logic, all rights reserved.

#ifndef DL_CORE_OS_H
#define DL_CORE_OS_H

#include "dl_math.h"
#include "dl_string.h"
#include "dl_hashmap.h"
#include "dl_platform.h"

namespace dl {
namespace os {

// Macros for capturing OS errors on OS calls. The upshot here is that you will use this by putting
// an entire OS call inside the DL_OS_CALL macro, like so:
// 
//    DL_OS_CALL( DWORD len = GetEnvironmentVariableW( pName, NULL, 0 ) );
//    if ( !len )
//    {
//        auto code = currentErrorCode();
//        if ( code != ERROR_ENVVAR_NOT_FOUND && code != ERROR_SUCCESS )
//        {
//            auto msg = currentErrorMessage();
//            if ( msg.isEmpty() )
//                msg = "(no specific error reported)";
//    
//            logError( "GetEnvironmentVariableW (first) call failed: %s", msg.buf() );
//        }
//    }
// 
// Internally, this is using GetLastError, errno, and/or dlerror() to find out what the OS thinks
// and formatting that for you to retrieve via the currentError functions below.
//
#if DL_IS_GCC || DL_IS_CLANG
    #define DL_CALLER_NS dl::String( __func__ )
    #define DL_CALLER    dl::String( __func__ ).fromLast( ':', false )
#else
    #define DL_CALLER_NS dl::String( __FUNCTION__ )
    #define DL_CALLER    dl::String( __FUNCTION__ ).fromLast( ':', false )
#endif

#define DL_OS_CALL(oscall) \
    dl::os::clearCurrentError(); oscall; dl::os::captureCurrentError(DL_CALLER);

struct DL_API ErrorInfo
{
    Int code;
    String caller;
    String message;
};

DL_API ErrorInfo    currentError();
DL_API Int          currentErrorCode();
DL_API String       currentErrorMessage();
DL_API String       currentErrorCaller();

// Probably do not need to use these.
//
DL_API void         setCurrentError( Int code, String caller, String msg );
DL_API void         captureCurrentError( String caller = String() );
DL_API void         clearCurrentError();

// Not sure if this works, I have no speakers on my dev machine.
//
DL_API void         beep();

// Message box helpers. These use tiny file dialog and are not yet well tested.
//
DL_API void         infoBox( String title, String message );
DL_API void         warningBox( String title, String message );
DL_API void         errorBox( String title, String message );

// More complex OS dialogs. Also using tiny file dialog, and not yet well tested.
//
enum DialogType
{
    DialogType_OkCancel,
    DialogType_YesNo,
    DialogType_YesNoCancel
};

DL_API Bool         questionBox( String title, String message, DialogType dialogType = DialogType_YesNo, Bool defaultResponse = false, Bool* cancelled = nullptr );
DL_API Bool         okCancelBox( String title, String message, Bool defaultResponse = false );
DL_API Bool         yesNoBox( String title, String message, Bool defaultResponse = false );
DL_API Bool         yesNoCancelBox( String title, String message, Bool defaultResponse = false, Bool* cancelled = nullptr );
DL_API String       inputBox( String prompt, String message, String defaultInput = "" );
DL_API Rgb          chooseColor( String title = "Choose Color", Rgb defaultRgb = Rgb::make( 0.5, 0.5, 0.5 ) );

// Get current locale ID, e.g. "en", "es", etc.
//
DL_API String       currentLocale();

// Get a localized item from a map<String, String>.
//
DL_API String       localizedItem( const ds::HashMap<String, String>& map, String lang = "auto" );

// Machine name.
//
DL_API String       machineName();

// Env vars.
//
DL_API StringVector env();
DL_API String       getEnv( String name );
DL_API Int          setEnv( String name, String value, Bool overwrite = true );
DL_API Int          putEnv( String value );
DL_API Int          unsetEnv( String name );

// Really just a thing for us, not gonna tell more.
//
DL_API Bool         devMode();

// Library functions.
//
DL_API DL_HMODULE   dlOpen( String path );
DL_API bool         dlClose( DL_HMODULE mod );
DL_API DL_HSYMBOL   dlSym( String symbol, DL_HMODULE mod = DL_DEFMODULE );

// Functions for starting applications. The provided environment will not be used, if the started
// application requires elevation.
//
DL_API Int          execV( String exe, StringVector args, Bool wait = false, StringVector env = os::env() );
DL_API String       quote( String path, Bool force = false );
DL_API String       unquote( String path );

// Flush messages. This only has code on win32, a basic Peek/Get/Translate/DispatchMessage loop.
//
DL_API void         doApplicationEvents();

// Open a file in the default application.
//
DL_API Bool         openPath( String path, String& error );

// See if we are running under Rosetta.
//
DL_API Bool         runningOnRosetta();

// Prevent OS napping.
//
DL_API void         disableAppNap();
DL_API void         enableAppNap();

// These are used by gui & cli.
//
DL_API void         enableHighPriority();
DL_API void         disableHighPriority();
DL_API bool         useHighPriority();

// Processor groups and thread affinity.
//
DL_API UInt32       numProcessorGroups();
DL_API UInt32       numLogicalProcessors( Int32 processorGroup = -1 );
DL_API bool         shouldSetThreadAffinity( UInt32 numThreadsToBeUsed );
DL_API void         setThreadAffinity( void* threadHandle, UInt32 affinity );
DL_API void         setThreadAffinity( unsigned long threadHandle, UInt32 affinity );

} // os
} // dl

#endif // DL_CORE_OS_H

